#!/bin/sh

STATUS="$SDO_ROOT_DIR"/var/lib/sdcard-overlay/disabled-logs.status
CONF="$SDO_ROOT_DIR"/etc/default/sdcard-overlay.conf
PROG=$(basename "$0" .sh)

# TODO
# - handle: dsm-firmware-upgrade, sdcard-overlay, pagekite
# - move statistics logfiles into dedicated folder
# - lighttpd does not like disabled logs
# - create dedicated logfolder for: dss11-websetup dss-addons
#   e.g. not a subfolder of dss

# these flags might be overriden when source CONF
CRITICAL_USAGE_PERCENT=80 # critical if usage higher than
CRITICAL_FREE_SPACE=60    # MB, needed for update

# anything that has dedicated folder in /var/log can be listed
# including 'dss-addons', that doesn't have a matching service
MANAGED_DAEMONS="ds485p vdsm dsa dss"

die()
{
    echo "ERROR: $1 ...  exiting"
    exit 1
}

fs_pcent_used()
{
    df -m $ROOTFS | tail -1 | awk '{print substr($5,1,length($5)-1)}'
}

fs_free_mb() {
    df -m $ROOTFS | tail -1 | awk '{print $4}'
}

enough_free_space() {
    local P=$(fs_pcent_used)
    local S=$(fs_free_mb)
    echo "used : ${P}% | free : ${S} MB"
    if [ $P -gt $CRITICAL_USAGE_PERCENT ] ||\
        [ $S -lt $CRITICAL_FREE_SPACE ] ||\
        [ -n "$FORCE_DISABLE" ] ; then
        return 1
    fi
    return 0
}

runit_service()
{
    test -d "$SDO_ROOT_DIR"/etc/runit/$1
}

initd_service()
{
    test -f "$SDO_ROOT_DIR"/etc/init.d/$1
}

is_runit_svlogd()
{
    # does it use svlogd?
    test -f "$SDO_ROOT_DIR"/etc/runit/$1/log/run
}

is_runit_svlogd_disabled()
{
    daemon=$1
    # true if no it use doesn't use svlogd
    test -d "$SDO_ROOT_DIR"/etc/runit/$daemon/log || return 1
    test -f "$SDO_ROOT_DIR"/etc/runit/$daemon/log/down
}

runit_svlogd()
{
    daemon=$1
    action=$2

    [ -n "$daemon" ] || die "runit_svlogd: missing daemon argument"
    [ -n "$action" ] || die "runit_svlogd: missing start/stop argument"

    is_runit_svlogd $daemon || return 0

    case $action in
        enable|start)
            echo "restore svlogd for $daemon"
            rm "$SDO_ROOT_DIR"/etc/runit/$daemon/log/down 2> /dev/null
            ;;
        disable|stop)
            # touching file 'down' tells runit to skip this service
            echo "disable svlogd for $daemon"
            touch "$SDO_ROOT_DIR"/etc/runit/$daemon/log/down
            ;;
        *)
            die "runit_svlogd: unknown action $action"
            ;;
    esac

    #
    # sv start/stop only applies to the managed process, e.g. 'dss'. But the
    # logging service, collecting all the output is not restarted. To totally
    # reload deamon config, we need to restart runsv control deamon.
    #
    # runsvdir:
    # it's a super-supervisor. Runsvdir runs every xx seconds, and restarts any
    # runsv service currently down, (no status pipe present). When runsv starts,
    # it inspects its configuration (down files), and brings up the service as
    # requested.
    #
    # NOTE:
    # 1) Make sure that runsvdir is the mother of all runsv processes. We then
    # can easily shutdown all runsv processes by sending runsvdir a signal,
    # which will then forward it to all its children. It will not terminate a
    # runsv processes we have started manually. This is important in case of
    # runit software upgrade. NEVER START RUNSV MANUALLY
    # 2) To apply configuration changes 'sv shutdown' must be used. THE RESTART
    # IS TRIGGERED ASYNCHRONOUSLY, OUTSIDE OF OUR CONTROL, this includes logging
    # if not perminantly disabled via 'down' file. Deal with it!
    # 3) consequence of 2): 'sv down' state is kept in memory of runsv, only
    # 'down' files are persistent. Upon runsv exit, temporary information is
    # lost. The daemon(dss) will restart automatically upon runsv restart
    #
    sv -w 30 force-shutdown $daemon || true
    # killall -9 $daemon || true # kill also those that escaped the watchdog
}

lookup_logdir()
{
    daemon=$1
    case $daemon in
        sdcard-overlay)   echo $SDO_ROOT_DIR/var/log/sdcard;;
        dss-addons)       echo $SDO_ROOT_DIR/var/log/dss/js;;
        *)                echo $SDO_ROOT_DIR/var/log/$daemon;;
    esac
}

do_daemon()
{
    D=$1
    action=$2

    case $action in
        start|stop);;
        *) die "do_daemon: invalid action $action";;
    esac

    case $D in
        dss-addons)
            # separate logdir part of dss
            return 0;;
        dss11-websetup)
            # separate logdir part of lighttpd
            return 0;;
        *);;
    esac

    if runit_service $D; then
        case $action in
            stop)  ACT="sv -w 30 force-stop $D";;
            start) ACT="sv start $D";;
        esac
    else
        # TODO: unused, all managed daemons use runit
        ACT="\"$SDO_ROOT_DIR\"/etc/init.d/$D $action"
    fi

    eval "${ACT}"
}

mount_ro_aufs()
{
    TARGET=$1

    # if fs is full, mkdir might fail and we can't create nullfs
    [ -d $TARGET ] || { echo "$1 does not exist"; return 2; }
    mount -t aufs -o br:$TARGET=ro none $TARGET
    [ $? -eq 0 ] || { echo "failed to mount overlay $1"; return 3; }
}

disable_logging()
{
    SERVICE=$1
    echo "disable logging for $SERVICE"

    if runit_service $SERVICE; then
        # The runsv daemon will be shutdown and restarted
        # asynchronously. But logging will be disabled, no
        # open open file descriptors. So never mind
        runit_svlogd $SERVICE stop
    fi

    LOGDIR=`lookup_logdir $SERVICE`
    [ -d "$LOGDIR" ] || { echo "no such folder $LOGDIR"; return 1; }

    # dss/js logfolder is not recreated if log/run, is not
    # started! Only delete top level files
    (cd "$LOGDIR"; for i in *; do [ -f $i ] && rm $i; done)

    # TODO: once the current implementation seems to be
    # stable, enable this
    # mount_ro_aufs $LOGDIR
}

select_victim()
{
    # make sure it exists
    mkdir -p "$(dirname "$STATUS")"

    echo "select victim $MANAGED_DAEMONS"
    for D in $MANAGED_DAEMONS; do
        if enough_free_space; then
            break
        fi

        LOGDIR=`lookup_logdir $D`
        if mountpoint -q $LOGDIR; then
            echo "$D already disabled(aufs)...  skip"
            continue
        fi

        if [ ! -d "$LOGDIR" ]; then
            echo "Unknown service $D, missing logfolder... skip"
            continue
        fi

        if is_runit_svlogd_disabled $D; then
            echo "$D already disabled(svlogd)... skip"
            continue
        fi

        do_daemon $D stop
        disable_logging $D
        echo $D >> $STATUS
        do_daemon $D start
    done
}

system_boot_apply_disabling()
{
    test -f "$STATUS" || {
        # nothing to do
        return 0
    }

    for D in `cat $STATUS`; do
        LOGDIR=`lookup_logdir $D`
        if mountpoint -q $LOGDIR; then
            continue
        fi

        if [ $D = "dss-addons" ]; then
            # dss/js does not exist if dss already mounted
            if mountpoint -q /var/log/dss; then
                # skip nested overlay
                continue
            fi
        fi

        disable_logging $D
    done
}

if [ -n "$UNITTEST" ]; then
    return
fi

#
# MAIN()
#
# WARNING: any command in here is not covered
# by the unit test, test them manually. At least
# copy&paste them into a shell
#

# allow to overwrite any variable
[ -f "$CONF" ] && . "$CONF"

echo
echo "conf: critical usage $CRITICAL_USAGE_PERCENT%"
echo "conf: minimal free space $CRITICAL_FREE_SPACE MB"
echo "conf: managed daemons $MANAGED_DAEMONS"
echo

usage()
{
    echo "$PROG [-bf]"
    echo "   disables logging of managed daemons if extern sd-card is not available"
    echo "   and internal flash usage is critical"
    echo ""
    echo "   Options:"
    echo "     -b     script is run at system boot"
    echo "     -f     force disabling of all managed deamons (requires sdcard not mounted)"
    echo ""
    echo "   FORCE OPTION"
    echo "     used for testing, example below"
    echo ""
    echo "     sdo-umount.sh"
    echo "     ${PROG} -f --> disable logging for all managed daemons"
    echo "     mount /media/card"
    echo "     ${PROG}    --> restores logging since sdcard is available"
    echo "     umount /media/card"
    echo "     ${PROG} -f --> another iteration"
    echo ""
    echo "     Reboot your target when done with testing, to ensure all services and"
    echo "     their logging services are in a known state"
    echo ""
    exit
}

while getopts ":bf" OPT; do
    #echo "foo $OPT -- $OPTARG"
    case $OPT in
        b) BOOTUP=1;;
        f) FORCE_DISABLE=1;;
        *) usage;;
    esac
done

restore_svlogd()
{
    OLD=`pwd`
    cd "$SDO_ROOT_DIR"/etc/runit
    for S in *; do
        test -f "$S"/log/down || continue
        echo "restore svlogd: $S"
        rm $S/log/down
    done
    cd $OLD
}

restore_corewatch()
{
    OLD=`pwd`
    for i in 2 3 4 5; do
        cd "$SDO_ROOT_DIR"/etc/rc$i.d/
        ln -s $SDO_ROOT_DIR/etc/init.d/corewatch S99corewatch
    done
    "$SDO_ROOT_DIR"/etc/init.d/corewatch start
    cd $OLD
}

if [ -n "$BOOTUP" ]; then
    # we run pretty early in the boot process
    # - runit not yet running
    # - sdcard-overlay already mounted, if card is okay

    if mountpoint -q /media/card; then
        # card is available re-enable logging
        rm $STATUS 2> /dev/null
        restore_svlogd
        restore_corewatch
    else
        echo "sdo-disable-logs: re-apply previous settings upon bootup"
        system_boot_apply_disabling
    fi
    exit 0
fi

if mountpoint -q /media/card; then
    # card is available nothing to do
    # not job of this script to cleanup internal flash
    echo "sd-card mounted, restore logging services"
    restore_svlogd
    # TODO runit_svlogd, reboot will fix
    return 0
fi

if enough_free_space; then
    echo "used : `fs_pcent_used`% | free : `fs_free_mb` MB"
    echo "enough space available, quit"
    return 0
fi

disable_corewatch()
{
    # burden on internal flash is too heavy
    # can't deinstall corewatch since task-dss-image depends on it
    "$SDO_ROOT_DIR"/etc/init.d/corewatch stop
    rm -f "$SDO_ROOT_DIR"/etc/rc*/S*corewatch
    rm -rf "$SDO_ROOT_DIR"/var/log/coredumps
}
disable_corewatch

select_victim

echo "used : `fs_pcent_used`% | free : `fs_free_mb` MB"

if ! enough_free_space; then
    echo "ERROR: failed to free enough disk space"
    return 1
fi

#
# TODO restore logging dirs, if enough space available (hysteresis)
#
